# coding: utf-8
"""
Object Permissions

There are three types of permissions for each object:
 - READ, being able to access all information on an object,
 - WRITE, being able to update an object, and
 - GRANT, being able to grant permissions to other users.

When an object is created, its creator initially gets GRANT permissions for the object. In addition, if the object
was created performing an instrument action, the instrument responsible users will have GRANT access rights, not as
individuals, but in their role of responsible users. So even when a user becomes responsible user AFTER an object was
created using an instrument, this user will then have GRANT rights, until he is removed from the list of responsible
users for the instrument.

Objects can be made public, which grants READ permissions to any logged-in user trying to access the object.
"""

import typing

import sqlalchemy

from .. import db
from . import errors
from . import actions
from .groups import get_user_groups, get_group_member_ids
from .instruments import get_instrument
from .notifications import create_notification_for_having_received_an_objects_permissions_request
from . import objects
from ..models import Permissions, UserObjectPermissions, GroupObjectPermissions, ProjectObjectPermissions, PublicObjects, Action, Object, DefaultUserPermissions, DefaultGroupPermissions, DefaultProjectPermissions, DefaultPublicPermissions
from . import projects
from . import settings
from .users import get_user, get_administrators


__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def object_is_public(object_id: int):
    return PublicObjects.query.filter_by(object_id=object_id).first() is not None


def set_object_public(object_id: int, is_public: bool = True):
    if not is_public:
        PublicObjects.query.filter_by(object_id=object_id).delete()
    elif not object_is_public(object_id):
        db.session.add(PublicObjects(object_id=object_id))
    db.session.commit()


def get_object_permissions_for_users(object_id: int, include_instrument_responsible_users: bool = True, include_groups: bool = True, include_projects: bool = True, include_readonly: bool = True, include_admin_permissions: bool = True):
    object_permissions = {}

    if include_admin_permissions:
        for user in get_administrators():
            if not settings.get_user_settings(user.id)['USE_ADMIN_PERMISSIONS']:
                # skip admins who do not use admin permissions
                continue
            if user.is_readonly:
                object_permissions[user.id] = Permissions.READ
            else:
                object_permissions[user.id] = Permissions.GRANT

    for user_object_permissions in UserObjectPermissions.query.filter_by(object_id=object_id).all():
        object_permissions[user_object_permissions.user_id] = user_object_permissions.permissions
    if include_instrument_responsible_users:
        for user_id in _get_object_responsible_user_ids(object_id):
            object_permissions[user_id] = Permissions.GRANT
    if include_groups:
        for group_object_permissions in GroupObjectPermissions.query.filter_by(object_id=object_id).all():
            for user_id in get_group_member_ids(group_object_permissions.group_id):
                if user_id not in object_permissions or object_permissions[user_id] in group_object_permissions.permissions:
                    object_permissions[user_id] = group_object_permissions.permissions
    if include_projects:
        for project_object_permissions in ProjectObjectPermissions.query.filter_by(object_id=object_id).all():
            for user_id, permissions in projects.get_project_member_user_ids_and_permissions(project_object_permissions.project_id, include_groups=include_groups).items():
                permissions = min(permissions, project_object_permissions.permissions)
                previous_permissions = object_permissions.get(user_id, Permissions.NONE)
                object_permissions[user_id] = max(previous_permissions, permissions)
    if include_readonly:
        for user_id in object_permissions:
            if get_user(user_id).is_readonly:
                object_permissions[user_id] = min(object_permissions[user_id], Permissions.READ)
    return object_permissions


def get_object_permissions_for_groups(object_id: int, include_projects: bool = False) -> typing.Dict[int, Permissions]:
    object_permissions = {}
    for group_object_permissions in GroupObjectPermissions.query.filter_by(object_id=object_id).all():
        if group_object_permissions.permissions != Permissions.NONE:
            object_permissions[group_object_permissions.group_id] = group_object_permissions.permissions
    if include_projects:
        for project_object_permissions in ProjectObjectPermissions.query.filter_by(object_id=object_id).all():
            for group_id, permissions in projects.get_project_member_group_ids_and_permissions(project_object_permissions.project_id).items():
                permissions = min(permissions, project_object_permissions.permissions)
                previous_permissions = object_permissions.get(group_id, Permissions.NONE)
                object_permissions[group_id] = max(previous_permissions, permissions)
    return object_permissions


def get_object_permissions_for_projects(object_id: int) -> typing.Dict[int, Permissions]:
    object_permissions = {}
    for project_object_permissions in ProjectObjectPermissions.query.filter_by(object_id=object_id).all():
        if project_object_permissions.permissions != Permissions.NONE:
            object_permissions[project_object_permissions.project_id] = project_object_permissions.permissions
    return object_permissions


def _get_object_responsible_user_ids(object_id):
    object = objects.get_object(object_id)
    try:
        action = actions.get_action(object.action_id)
    except errors.ActionDoesNotExistError:
        return []
    if action.instrument_id is None:
        return []
    instrument = get_instrument(action.instrument_id)
    return [user.id for user in instrument.responsible_users]


def get_user_object_permissions(object_id: int, user_id: int, include_instrument_responsible_users: bool = True, include_groups: bool = True, include_projects: bool = True, include_readonly: bool = True, include_admin_permissions: bool = True):
    user = get_user(user_id)

    if include_admin_permissions:
        # administrators have GRANT permissions if they use admin permissions
        if user.is_admin and settings.get_user_settings(user.id)['USE_ADMIN_PERMISSIONS']:
            # unless they are limited to READ permissions
            if user.is_readonly:
                return Permissions.READ
            else:
                return Permissions.GRANT

    if include_instrument_responsible_users and include_groups and include_projects:
        stmt = db.text("""
        SELECT
        MAX(permissions_int)
        FROM user_object_permissions_by_all
        WHERE (user_id = :user_id OR user_id IS NULL) AND object_id = :object_id
        """)
        permissions_int = db.engine.execute(stmt, {
            'user_id': user_id,
            'object_id': object_id
        }).fetchone()[0]
        if permissions_int is None or permissions_int <= 0:
            return Permissions.NONE
        elif include_readonly and user.is_readonly and permissions_int in (1, 2, 3):
            return Permissions.READ
        elif permissions_int == 1:
            return Permissions.READ
        elif permissions_int == 2:
            return Permissions.WRITE
        elif permissions_int >= 3:
            return Permissions.GRANT

    if include_instrument_responsible_users:
        # instrument responsible users always have GRANT permissions for an object
        if user_id in _get_object_responsible_user_ids(object_id):
            if include_readonly and user.is_readonly:
                return Permissions.READ
            return Permissions.GRANT
    # other users might have been granted permissions, either individually or as group or project members
    user_object_permissions = UserObjectPermissions.query.filter_by(object_id=object_id, user_id=user_id).first()
    if user_object_permissions is None:
        permissions = Permissions.NONE
    else:
        permissions = user_object_permissions.permissions
    if include_readonly and user.is_readonly and Permissions.READ in permissions:
        return Permissions.READ
    if Permissions.GRANT in permissions:
        return permissions
    if include_groups:
        for group in get_user_groups(user_id):
            group_object_permissions = GroupObjectPermissions.query.filter_by(object_id=object_id, group_id=group.id).first()
            if group_object_permissions is not None and permissions in group_object_permissions.permissions:
                permissions = group_object_permissions.permissions
    if include_readonly and user.is_readonly and Permissions.READ in permissions:
        return Permissions.READ
    if Permissions.GRANT in permissions:
        return permissions
    if include_projects:
        for user_project in projects.get_user_projects(user_id, include_groups=include_groups):
            user_project_permissions = projects.get_user_project_permissions(user_project.id, user_id, include_groups=include_groups)
            if user_project_permissions not in permissions:
                project_object_permissions = ProjectObjectPermissions.query.filter_by(object_id=object_id, project_id=user_project.id).first()
                if project_object_permissions is not None:
                    permissions = min(user_project_permissions, project_object_permissions.permissions)
    if include_readonly and user.is_readonly and Permissions.READ in permissions:
        return Permissions.READ
    if Permissions.READ in permissions:
        return permissions
    # lastly, the object may be public, so all users have READ permissions
    if object_is_public(object_id):
        return Permissions.READ
    # otherwise the user has no permissions for this object
    return Permissions.NONE


def set_user_object_permissions(object_id: int, user_id: int, permissions: Permissions):
    assert user_id is not None
    if permissions == Permissions.NONE:
        UserObjectPermissions.query.filter_by(object_id=object_id, user_id=user_id).delete()
    else:
        user_object_permissions = UserObjectPermissions.query.filter_by(object_id=object_id, user_id=user_id).first()
        if user_object_permissions is None:
            user_object_permissions = UserObjectPermissions(user_id=user_id, object_id=object_id, permissions=permissions)
        else:
            user_object_permissions.permissions = permissions
        db.session.add(user_object_permissions)
    db.session.commit()


def set_group_object_permissions(object_id: int, group_id: int, permissions: Permissions):
    assert group_id is not None
    if permissions == Permissions.NONE:
        GroupObjectPermissions.query.filter_by(object_id=object_id, group_id=group_id).delete()
    else:
        group_object_permissions = GroupObjectPermissions.query.filter_by(object_id=object_id, group_id=group_id).first()
        if group_object_permissions is None:
            group_object_permissions = GroupObjectPermissions(object_id=object_id, group_id=group_id, permissions=permissions)
        else:
            group_object_permissions.permissions = permissions
        db.session.add(group_object_permissions)
    db.session.commit()


def set_project_object_permissions(object_id: int, project_id: int, permissions: Permissions) -> None:
    """
    Sets the permissions for a project for an object. Clears the permissions
    if called with Permissions.NONE.

    :param object_id: the object ID of an existing object
    :param project_id: the project ID of an existing probject
    :param permissions: the new permissions
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.ProjectDoesNotExistError: when no project with the given
        project ID exists
    """
    projects.get_project(project_id)
    if permissions == Permissions.NONE:
        ProjectObjectPermissions.query.filter_by(object_id=object_id, project_id=project_id).delete()
    else:
        project_object_permissions = ProjectObjectPermissions.query.filter_by(object_id=object_id, project_id=project_id).first()
        if project_object_permissions is None:
            project_object_permissions = ProjectObjectPermissions(project_id=project_id, object_id=object_id, permissions=permissions)
        else:
            project_object_permissions.permissions = permissions
        db.session.add(project_object_permissions)
    db.session.commit()


def set_initial_permissions(obj):
    default_user_permissions = get_default_permissions_for_users(creator_id=obj.user_id)
    for user_id, permissions in default_user_permissions.items():
        set_user_object_permissions(object_id=obj.object_id, user_id=user_id, permissions=permissions)
    default_group_permissions = get_default_permissions_for_groups(creator_id=obj.user_id)
    for group_id, permissions in default_group_permissions.items():
        set_group_object_permissions(object_id=obj.object_id, group_id=group_id, permissions=permissions)
    default_project_permissions = get_default_permissions_for_projects(creator_id=obj.user_id)
    for project_id, permissions in default_project_permissions.items():
        set_project_object_permissions(object_id=obj.object_id, project_id=project_id, permissions=permissions)
    should_be_public = default_is_public(creator_id=obj.user_id)
    set_object_public(object_id=obj.object_id, is_public=should_be_public)


def get_object_info_with_permissions(
        user_id: int,
        permissions: Permissions,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        action_id: typing.Optional[int] = None,
        action_type_id: typing.Optional[int] = None,
        object_ids: typing.Optional[typing.Sequence[int]] = None
) -> typing.List[Object]:

    user = get_user(user_id)

    # readonly users may not have more than READ permissions
    if user.is_readonly and permissions != Permissions.READ:
        return []

    if action_type_id is not None and action_id is not None:
        action_filter = db.and_(Action.type_id == action_type_id, Action.id == action_id)
    elif action_type_id is not None:
        action_filter = (Action.type_id == action_type_id)
    elif action_id is not None:
        action_filter = (Action.id == action_id)
    else:
        action_filter = None

    if user.is_admin and settings.get_user_settings(user.id)['USE_ADMIN_PERMISSIONS']:
        # admins who use admin permissions do not need permission-based filtering
        stmt = db.text("""
        SELECT
        o.object_id, o.data -> 'name' ->> 'text' as name_text, o.action_id, 3 as max_permission
        FROM objects_current AS o
        """)
    else:
        stmt = db.text("""
        SELECT
        o.object_id, o.data -> 'name' ->> 'text' as name_text, o.action_id, p.max_permission
        FROM (
            SELECT
            object_id, MAX(permissions_int) AS max_permission
            FROM user_object_permissions_by_all
            WHERE user_id = :user_id OR user_id IS NULL
            GROUP BY (object_id)
            HAVING MAX(permissions_int) >= :min_permissions_int
        ) AS p
        JOIN objects_current AS o ON o.object_id = p.object_id
        """)

    stmt = stmt.columns(
        objects.Objects._current_table.c.object_id,
        sqlalchemy.sql.expression.column('name_text'),
        objects.Objects._current_table.c.action_id,
        sqlalchemy.sql.expression.column('max_permission'),
    )

    parameters = {
        'min_permissions_int': permissions.value,
        'user_id': user_id
    }

    table = stmt.alias('tbl')

    where = table.c.action_id.in_(db.select([Action.__table__.c.id], whereclause=action_filter)) if action_filter is not None else None

    if object_ids is not None:
        object_id_where = table.c.object_id.in_(tuple(object_ids))
        if where is None:
            where = object_id_where
        else:
            where = db.and_(where, object_id_where)

    table = db.select(
        columns=[table],
        whereclause=where
    ).order_by(db.desc(table.c.object_id)).limit(limit).offset(offset)

    object_infos = db.session.execute(table, parameters).fetchall()

    return object_infos


def get_objects_with_permissions(
        user_id: int,
        permissions: Permissions,
        filter_func: typing.Callable = lambda data: True,
        sorting_func: typing.Optional[typing.Callable[[typing.Any], typing.Any]] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        action_id: typing.Optional[int] = None,
        action_type_id: typing.Optional[int] = None,
        other_user_id: typing.Optional[int] = None,
        other_user_permissions: typing.Optional[Permissions] = None,
        group_id: typing.Optional[int] = None,
        group_permissions: typing.Optional[Permissions] = None,
        project_id: typing.Optional[int] = None,
        project_permissions: typing.Optional[Permissions] = None,
        object_ids: typing.Optional[typing.Sequence[int]] = None,
        num_objects_found: typing.Optional[typing.List[int]] = None,
        name_only: bool = False,
        **kwargs
) -> typing.List[Object]:

    user = get_user(user_id)

    # readonly users may not have more than READ permissions
    if user.is_readonly and permissions != Permissions.READ:
        return []

    if action_type_id is not None and action_id is not None:
        action_filter = db.and_(Action.type_id == action_type_id, Action.id == action_id)
    elif action_type_id is not None:
        action_filter = (Action.type_id == action_type_id)
    elif action_id is not None:
        action_filter = (Action.id == action_id)
    else:
        action_filter = None

    parameters = {
        'min_permissions_int': permissions.value,
        'user_id': user_id
    }

    if name_only:
        stmt = """
        SELECT
        o.object_id, o.version_id, o.action_id, jsonb_set('{"name": {"_type": "text", "text": ""}}', '{name,text}', o.data -> 'name' -> 'text') as data, '{"title": "Object", "type": "object", "properties": {"name": {"title": "Name", "type": "text"}}}'::jsonb as schema, o.user_id, o.utc_datetime
        FROM objects_current AS o
        """
    else:
        stmt = """
        SELECT
        o.object_id, o.version_id, o.action_id, o.data, o.schema, o.user_id, o.utc_datetime
        FROM objects_current AS o
        """

    if not user.is_admin or not settings.get_user_settings(user.id)['USE_ADMIN_PERMISSIONS']:
        stmt += """
        JOIN (
            SELECT
            u.object_id
            FROM user_object_permissions_by_all as u
            WHERE u.user_id = :user_id OR u.user_id IS NULL
            GROUP BY (u.object_id)
            HAVING MAX(u.permissions_int) >= :min_permissions_int
        ) AS up
        ON o.object_id = up.object_id
        """

    if other_user_id is not None:
        stmt += """
        JOIN (
            SELECT
            u.object_id
            FROM user_object_permissions_by_all as u
            WHERE u.user_id = :other_user_id OR u.user_id IS NULL
            GROUP BY (u.object_id)
            HAVING MAX(u.permissions_int) >= :min_other_user_permissions_int
        ) AS oup
        ON o.object_id = oup.object_id
        """
        parameters['other_user_id'] = other_user_id
        if other_user_permissions is None:
            other_user_permissions = permissions
        parameters['min_other_user_permissions_int'] = other_user_permissions.value

    if project_id is not None:
        stmt += """
        JOIN project_object_permissions as pp
        ON (
            pp.object_id = o.object_id AND
            pp.project_id = :project_id AND
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> pp.permissions::text)::int >= :min_project_permissions_int
        )
        """
        parameters['project_id'] = project_id
        if project_permissions is None:
            project_permissions = permissions
        parameters['min_project_permissions_int'] = project_permissions.value

    if group_id is not None:
        stmt += """
        JOIN group_object_permissions as gp
        ON (
            gp.object_id = o.object_id AND
            gp.group_id = :group_id AND
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> gp.permissions::text)::int >= :min_group_permissions_int
        )
        """
        parameters['group_id'] = group_id
        if group_permissions is None:
            group_permissions = permissions
        parameters['min_group_permissions_int'] = group_permissions.value

    if object_ids:
        stmt += """
        WHERE o.object_id IN :object_ids
        """
        parameters['object_ids'] = tuple(object_ids)

    table = sqlalchemy.sql.alias(db.text(stmt).columns(
        objects.Objects._current_table.c.object_id,
        objects.Objects._current_table.c.version_id,
        objects.Objects._current_table.c.action_id,
        objects.Objects._current_table.c.data,
        objects.Objects._current_table.c.schema,
        objects.Objects._current_table.c.user_id,
        objects.Objects._current_table.c.utc_datetime
    ))

    return objects.get_objects(
        filter_func=filter_func,
        action_filter=action_filter,
        table=table,
        parameters=parameters,
        sorting_func=sorting_func,
        limit=limit,
        offset=offset,
        num_objects_found=num_objects_found,
        **kwargs
    )


class InvalidDefaultPermissionsError(Exception):
    pass


def get_default_permissions_for_users(creator_id: int) -> typing.Dict[int, Permissions]:
    default_permissions = {}
    for user_permissions in DefaultUserPermissions.query.filter_by(creator_id=creator_id).all():
        if user_permissions.permissions != Permissions.NONE:
            default_permissions[user_permissions.user_id] = user_permissions.permissions
    default_permissions[creator_id] = Permissions.GRANT
    return default_permissions


def set_default_permissions_for_user(creator_id: int, user_id: int, permissions: Permissions) -> None:
    if user_id == creator_id:
        if permissions == Permissions.GRANT:
            return
        else:
            raise InvalidDefaultPermissionsError("creator will always get GRANT permissions")
    user_permissions = DefaultUserPermissions.query.filter_by(creator_id=creator_id, user_id=user_id).first()
    if user_permissions is None:
        user_permissions = DefaultUserPermissions(creator_id=creator_id, user_id=user_id, permissions=permissions)
    else:
        user_permissions.permissions = permissions
    db.session.add(user_permissions)
    db.session.commit()


def get_default_permissions_for_groups(creator_id: int) -> typing.Dict[int, Permissions]:
    default_permissions = {}
    for group_permissions in DefaultGroupPermissions.query.filter_by(creator_id=creator_id).all():
        if group_permissions.permissions != Permissions.NONE:
            default_permissions[group_permissions.group_id] = group_permissions.permissions
    return default_permissions


def set_default_permissions_for_group(creator_id: int, group_id: int, permissions: Permissions) -> None:
    group_permissions = DefaultGroupPermissions.query.filter_by(creator_id=creator_id, group_id=group_id).first()
    if group_permissions is None:
        group_permissions = DefaultGroupPermissions(creator_id=creator_id, group_id=group_id, permissions=permissions)
    else:
        group_permissions.permissions = permissions
    db.session.add(group_permissions)
    db.session.commit()


def get_default_permissions_for_projects(creator_id: int) -> typing.Dict[int, Permissions]:
    default_permissions = {}
    for project_permissions in DefaultProjectPermissions.query.filter_by(creator_id=creator_id).all():
        if project_permissions.permissions != Permissions.NONE:
            default_permissions[project_permissions.project_id] = project_permissions.permissions
    return default_permissions


def set_default_permissions_for_project(creator_id: int, project_id: int, permissions: Permissions) -> None:
    project_permissions = DefaultProjectPermissions.query.filter_by(creator_id=creator_id, project_id=project_id).first()
    if project_permissions is None:
        project_permissions = DefaultProjectPermissions(creator_id=creator_id, project_id=project_id, permissions=permissions)
    else:
        project_permissions.permissions = permissions
    db.session.add(project_permissions)
    db.session.commit()


def default_is_public(creator_id: int) -> bool:
    public_permissions = DefaultPublicPermissions.query.filter_by(creator_id=creator_id).first()
    if public_permissions is None:
        return False
    return public_permissions.is_public


def set_default_public(creator_id: int, is_public: bool = True) -> None:
    public_permissions = DefaultPublicPermissions.query.filter_by(creator_id=creator_id).first()
    if public_permissions is None:
        public_permissions = DefaultPublicPermissions(creator_id=creator_id, is_public=is_public)
    else:
        public_permissions.is_public = is_public
    db.session.add(public_permissions)
    db.session.commit()


def request_object_permissions(requester_id: int, object_id: int) -> None:
    permissions_by_user = get_object_permissions_for_users(object_id)
    if Permissions.READ in permissions_by_user.get(requester_id, Permissions.NONE):
        return
    granting_user_ids = [
        user_id
        for user_id, permissions in permissions_by_user.items()
        if Permissions.GRANT in permissions
    ]
    for user_id in granting_user_ids:
        create_notification_for_having_received_an_objects_permissions_request(user_id, object_id, requester_id)


def copy_permissions(target_object_id: int, source_object_id: int) -> None:
    PublicObjects.query.filter_by(object_id=target_object_id).delete()
    UserObjectPermissions.query.filter_by(object_id=target_object_id).delete()
    GroupObjectPermissions.query.filter_by(object_id=target_object_id).delete()
    ProjectObjectPermissions.query.filter_by(object_id=target_object_id).delete()

    if PublicObjects.query.filter_by(object_id=source_object_id).first() is not None:
        db.session.add(PublicObjects(object_id=target_object_id))
    for user_object_permissions in UserObjectPermissions.query.filter_by(object_id=source_object_id).all():
        db.session.add(UserObjectPermissions(
            object_id=target_object_id,
            user_id=user_object_permissions.user_id,
            permissions=user_object_permissions.permissions
        ))
    for group_object_permissions in GroupObjectPermissions.query.filter_by(object_id=source_object_id).all():
        db.session.add(GroupObjectPermissions(
            object_id=target_object_id,
            group_id=group_object_permissions.group_id,
            permissions=group_object_permissions.permissions
        ))
    for project_object_permissions in ProjectObjectPermissions.query.filter_by(object_id=source_object_id).all():
        db.session.add(ProjectObjectPermissions(
            object_id=target_object_id,
            project_id=project_object_permissions.project_id,
            permissions=project_object_permissions.permissions
        ))
    db.session.commit()



# coding: utf-8
"""

"""

import pint

from ..units import ureg
from ..errors import UndefinedUnitError

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def units_are_valid(units: str) -> bool:
    try:
        ureg.Unit(units)
        return True
    except pint.UndefinedUnitError:
        return False


def get_dimensionality_for_units(units: str) -> str:
    """
    Returns the units' dimensionality in the dimensionality syntax of the pint package.

    :param units: a valid
    :return: dimensionality as string
    :raise errors.UndefinedUnitError: if the units are undefined
    """
    try:
        return str(ureg.Unit(units).dimensionality)
    except pint.UndefinedUnitError:
        raise UndefinedUnitError()



# coding: utf-8
"""

"""

import flask
import flask_login

from .. import frontend
from ...logic.authentication import login
from ...logic.users import get_user
from ...frontend.users_forms import SigninForm, SignoutForm
from ... import login_manager


@login_manager.user_loader
def load_user(user_id):
    try:
        user_id = int(user_id)
    except ValueError:
        return None
    user = get_user(user_id)
    if not user.is_active:
        flask.flash('Your user account has been deactivated. Please contact an administrator.', 'error')
        return None
    return user


def _is_url_safe_for_redirect(url: str) -> bool:
    """
    Checks whether a URL is safe for redirecting a user to it.

    URLs are deemed safe if they are absolute URLs on the current domain and
    server path. This is enforced by only allowing URLs that start with the
    absolute server path and only consist of alphanumerical characters or the
    following characters: /=?&_.+-

    :param url: the URL to check
    :return: whether it is safe to redirect to or not
    """
    server_path = flask.current_app.config.get('SERVER_PATH', '/')
    # ensure the server path starts with a / to avoid relative paths or non-local paths
    if not server_path.startswith('/'):
        server_path = '/' + server_path
    # ensure the server path ends with a / to avoid paths with it as a prefix
    if not server_path.endswith('/'):
        server_path = server_path + '/'
    # prevent double slashes that would change the domain
    if url.startswith('//'):
        return False
    return url.startswith(server_path) and all(c in '/=?&_.+-' or c.isalnum() for c in url)


def _redirect_to_next_url():
    next_url = flask.request.args.get('next', flask.url_for('.index'))
    index_url = flask.url_for('.index')
    if not _is_url_safe_for_redirect(next_url):
        next_url = index_url
    return flask.redirect(next_url)


def _sign_in_impl(is_for_refresh):
    form = SigninForm()
    has_errors = False
    if form.validate_on_submit():
        username = form.username.data
        # enforce lowercase for username to ensure case insensitivity
        # TODO: move this to the model and logic layer
        username = username.lower()
        password = form.password.data
        user = login(username, password)
        if user and not user.is_active:
            flask_login.logout_user()
            flask.flash('Your user account has been deactivated. Please contact an administrator.', 'error')
            return flask.redirect(flask.url_for('frontend.sign_in'))
        if is_for_refresh:
            if user and user == flask_login.current_user:
                flask_login.confirm_login()
                return _redirect_to_next_url()
        else:
            if user:
                remember_me = form.remember_me.data
                flask_login.login_user(user, remember=remember_me)
                if user.is_admin and password == 'password':
                    flask.flash(
                        'You are using the default admin password from the '
                        'SampleDB documentation. Please change your password '
                        'before making this SampleDB instance available to '
                        'other users.',
                        'warning'
                    )
                return _redirect_to_next_url()
        has_errors = True
    elif form.errors:
        has_errors = True
    return flask.render_template('sign_in.html', form=form, is_for_refresh=is_for_refresh, has_errors=has_errors)


@frontend.route('/users/me/sign_in', methods=['GET', 'POST'])
def sign_in():
    if flask_login.current_user.is_authenticated:
        # if the user was already authenticated, redirect to the index.
        return flask.redirect(flask.url_for('.index'))
    return _sign_in_impl(is_for_refresh=False)


@frontend.route('/users/me/refresh_sign_in', methods=['GET', 'POST'])
@flask_login.login_required
def refresh_sign_in():
    if flask_login.current_user.is_authenticated and flask_login.login_fresh():
        # if the login was already fresh, redirect to the next_url or index.
        return _redirect_to_next_url()
    return _sign_in_impl(is_for_refresh=True)


@frontend.route('/users/me/sign_out', methods=['GET', 'POST'])
@flask_login.login_required
def sign_out():
    form = SignoutForm()
    if form.validate_on_submit():
        flask_login.logout_user()
        return flask.redirect(flask.url_for('.index'))
    return flask.render_template('sign_out.html')



# coding: utf-8
"""

"""

import re

from flask_wtf import FlaskForm
from wtforms import StringField, RadioField, PasswordField, SubmitField, IntegerField
from wtforms.validators import DataRequired, Length, Email, ValidationError


class NewUserForm(FlaskForm):
    name = StringField('Full Name', validators=[DataRequired()])
    email = StringField('Contact Email', validators=[DataRequired("Please enter the contact email."), Email("Please enter your contact email.")])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=3)])
    authentication_method = RadioField('Authentication Method', choices=[('E', 'Email'), ('O', 'Other')], default='O')
    login = StringField('Login', validators=[DataRequired()])
    type = RadioField('Usertype', choices=[('P', 'Person'), ('O', 'Other')], default='O')
    admin = RadioField('Admin', choices=[('0', 'No'), ('1', 'Yes')], default='0')
    submit = SubmitField('Register')


class ChangeUserForm(FlaskForm):
    name = StringField('Full Name', validators=[DataRequired(), Length(min=1)])
    email = StringField('Contact Email', validators=[DataRequired("Please enter the contact email."), Email("Please enter your contact email.")])
    orcid = StringField('ORCID iD')
    affiliation = StringField('Affiliation')
    submit = SubmitField('Change Settings')

    def __init_(self, name=None, email=None):
        super(ChangeUserForm, self).__init__()

    def validate_orcid(self, field):
        orcid = field.data
        # accept empty ORCID iDs
        if orcid is None:
            return
        orcid = orcid.strip()
        if not orcid:
            return
        # accept full ORCID iDs
        orcid_prefix = 'https://orcid.org/'
        if orcid.startswith(orcid_prefix):
            orcid = orcid[len(orcid_prefix):]
        # check ORCID iD syntax
        if not re.fullmatch(r'\d{4}-\d{4}-\d{4}-\d{4}', orcid, flags=re.ASCII):
            raise ValidationError("Please enter a valid ORCID iD.")
        # keep sanitized ORCID iD on success
        field.data = orcid


class LoginForm(FlaskForm):
    username = StringField('Login', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=3)])
    submit = SubmitField('Login')


class AuthenticationForm(FlaskForm):
    login = StringField('Login', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=3)])
    authentication_method = RadioField('Authentication Method', choices=[('E', 'Email'), ('L', 'LDAP')], default='E')
    submit = SubmitField('Login')


class AuthenticationMethodForm(FlaskForm):
    id = IntegerField('Authentication_method_id', validators=[DataRequired()])
    submit = SubmitField('Remove')



# coding: utf-8
"""

"""

import flask
import flask_login

from .. import frontend


@frontend.route('/users/me/activity')
@flask_login.login_required
def current_user_activity():
    return flask.redirect(flask.url_for('.user_profile', user_id=flask_login.current_user.id))


@frontend.route('/users/<int:user_id>/activity')
@flask_login.login_required
def user_activity(user_id):
    return flask.redirect(flask.url_for('.user_profile', user_id=user_id))



# coding: utf-8
"""
Export all SampleDB information readable by the user to an archive.
"""

import datetime
import io
import json
import os
import tarfile
import time
import typing
import zipfile

import flask

from .. import logic


def get_archive_files(user_id: int, object_ids: typing.Optional[typing.List[int]] = None) -> typing.Dict[str, bytes]:
    archive_files = {}
    if object_ids is None:
        relevant_instrument_ids = {
            instrument.id
            for instrument in logic.instruments.get_instruments()
        }
    else:
        relevant_instrument_ids = set()
    relevant_action_ids = set()
    relevant_user_ids = set()
    relevant_location_ids = set()
    relevant_markdown_images = set()
    objects = logic.object_permissions.get_objects_with_permissions(
        user_id,
        logic.object_permissions.Permissions.READ
    )
    infos = {}
    object_infos = []
    for object in objects:
        if object_ids is not None and object.id not in object_ids:
            continue

        archive_files[f"sampledb_export/{object.id}.rdf"] = logic.rdf.generate_rdf(user_id, object.id)

        object_infos.append({
            'id': object.id,
            'action_id': object.action_id,
            'versions': [],
            'comments': [],
            'location_assignments': [],
            'publications': [],
            'files': []
        })
        relevant_action_ids.add(object.action_id)

        for object_version in logic.objects.get_object_versions(object.id):
            relevant_user_ids.add(object_version.user_id)
            object_infos[-1]['versions'].append(
                {
                    'id': object_version.version_id,
                    'user_id': object_version.user_id,
                    'utc_datetime': object_version.utc_datetime.isoformat(),
                    'schema': object_version.schema,
                    'data': object_version.data
                }
            )

            for referenced_user_id, _ in logic.objects.find_user_references(object_version, False):
                relevant_user_ids.add(referenced_user_id)

        for comment in logic.comments.get_comments_for_object(object.id):
            relevant_user_ids.add(comment.user_id)
            object_infos[-1]['comments'].append({
                'id': comment.id,
                'author_id': comment.user_id,
                'content': comment.content,
                'utc_datetime': comment.utc_datetime.isoformat()
            })

        for location_assignment in logic.locations.get_object_location_assignments(object.id):
            relevant_user_ids.add(location_assignment.user_id)
            relevant_user_ids.add(location_assignment.responsible_user_id)
            relevant_location_ids.add(location_assignment.location_id)
            object_infos[-1]['location_assignments'].append({
                'id': location_assignment.id,
                'assigning_user_id': location_assignment.user_id,
                'responsible_user_id': location_assignment.responsible_user_id,
                'location_id': location_assignment.location_id,
                'utc_datetime': location_assignment.utc_datetime.isoformat(),
            })

        log_entries = logic.object_log.get_object_log_entries(object.id, user_id)
        publication_log_entries = {
            (log_entry.data['doi'], log_entry.data['title']): log_entry
            for log_entry in log_entries
            if log_entry.type == logic.object_log.ObjectLogEntryType.LINK_PUBLICATION
        }
        for publication_info in logic.publications.get_publications_for_object(object.id):
            object_infos[-1]['publications'].append({
                'doi': publication_info.doi,
                'title': publication_info.title,
                'object_name': publication_info.object_name
            })
            if (publication_info.doi, publication_info.title) in publication_log_entries:
                publication_log_entry = publication_log_entries[(publication_info.doi, publication_info.title)]
                object_infos[-1]['publications'][-1]['user_id'] = publication_log_entry.user_id
                object_infos[-1]['publications'][-1]['utc_datetime'] = publication_log_entry.utc_datetime.isoformat()

        for file_info in logic.files.get_files_for_object(object.id):
            if not file_info.is_hidden:
                relevant_user_ids.add(file_info.user_id)
                object_infos[-1]['files'].append({
                    'id': file_info.id,
                    'hidden': False,
                    'title': file_info.title,
                    'description': file_info.description,
                    'uploader_id': file_info.user_id,
                    'utc_datetime': file_info.utc_datetime.isoformat()
                })
                if file_info.storage == 'local':
                    object_infos[-1]['files'][-1]['original_file_name'] = file_info.original_file_name
                    try:
                        file_bytes = file_info.open(read_only=True).read()
                    except Exception:
                        pass
                    else:
                        file_name = os.path.basename(file_info.original_file_name)
                        object_infos[-1]['files'][-1]['path'] = f'objects/{object.id}/files/{file_info.id}/{file_name}'
                        archive_files[f"sampledb_export/objects/{object.id}/files/{file_info.id}/{file_name}"] = file_bytes
                elif file_info.storage == 'url':
                    object_infos[-1]['files'][-1]['url'] = file_info.url
            else:
                object_infos[-1]['files'].append({
                    'id': file_info.id,
                    'hidden': True,
                    'hide_reason': file_info.hide_reason
                })
    infos['objects'] = object_infos

    action_infos = []
    for action_info in logic.actions.get_actions():
        if action_info.id in relevant_action_ids:
            relevant_user_ids.add(action_info.user_id)
            relevant_instrument_ids.add(action_info.instrument_id)
            action_permissions = logic.action_permissions.get_user_action_permissions(action_info.id, user_id)
            if logic.action_permissions.Permissions.READ in action_permissions:
                action_infos.append({
                    'id': action_info.id,
                    'type': action_info.type.object_name.lower(),
                    'name': action_info.name,
                    'user_id': action_info.user_id,
                    'instrument_id': action_info.instrument_id,
                    'description': action_info.description,
                    'description_as_html': action_info.description_as_html
                })
            if action_info.description_as_html:
                relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(action_info.description_as_html))
    infos['actions'] = action_infos

    instrument_infos = []
    for instrument_info in logic.instruments.get_instruments():
        if instrument_info.id in relevant_instrument_ids:
            relevant_user_ids.update({user.id for user in instrument_info.responsible_users})
            instrument_infos.append({
                'id': instrument_info.id,
                'name': instrument_info.name,
                'description': instrument_info.description,
                'description_as_html': instrument_info.description_as_html,
                'instrument_scientist_ids': [user.id for user in instrument_info.responsible_users],
                'instrument_log_entries': []
            })
            user_is_instrument_responsible = user_id in [user.id for user in instrument_info.responsible_users]
            if user_is_instrument_responsible:
                instrument_infos[-1]['notes'] = instrument_info.notes
                instrument_infos[-1]['notes_as_html'] = instrument_info.notes_as_html
            if instrument_info.users_can_view_log_entries or user_is_instrument_responsible:
                for log_entry in logic.instrument_log_entries.get_instrument_log_entries(instrument_info.id):
                    relevant_user_ids.add(log_entry.user_id)
                    instrument_infos[-1]['instrument_log_entries'].append({
                        'id': log_entry.id,
                        'author_id': log_entry.user_id,
                        'file_attachments': [],
                        'object_attachments': [],
                        'versions': []
                    })
                    for version in log_entry.versions:
                        instrument_infos[-1]['instrument_log_entries'][-1]['versions'].append({
                            'log_entry_id': log_entry.id,
                            'version_id': version.version_id,
                            'content': version.content,
                            'utc_datetime': version.utc_datetime.isoformat(),
                            'categories': []
                        })
                        for category in version.categories:
                            instrument_infos[-1]['instrument_log_entries'][-1]['versions'][-1]['categories'].append({
                                'id': category.id,
                                'title': category.title
                            })
                    file_attachments = logic.instrument_log_entries.get_instrument_log_file_attachments(log_entry.id)
                    for file_attachment in file_attachments:
                        file_name = os.path.basename(file_attachment.file_name)
                        file_path = f'instruments/{instrument_info.id}/log_entries/{log_entry.id}/files/{file_attachment.id}/{file_name}'
                        instrument_infos[-1]['instrument_log_entries'][-1]['file_attachments'].append({
                            'file_name': file_attachment.file_name,
                            'path': file_path
                        })
                        archive_files["sampledb_export/" + file_path] = file_attachment.content

                    object_attachments = logic.instrument_log_entries.get_instrument_log_object_attachments(log_entry.id)
                    for object_attachment in object_attachments:
                        instrument_infos[-1]['instrument_log_entries'][-1]['object_attachments'].append({
                            'object_id': object_attachment.object_id
                        })
            if instrument_info.description_as_html:
                relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(instrument_info.description_as_html))
            if user_is_instrument_responsible and instrument_info.notes_as_html:
                relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(instrument_info.notes_as_html))
    infos['instruments'] = instrument_infos

    user_infos = []
    for user_info in logic.users.get_users(exclude_hidden=False):
        if user_info.id in relevant_user_ids:
            user_infos.append({
                'id': user_info.id,
                'name': user_info.name,
                'orcid_id': f'https://orcid.org/{user_info.orcid}' if user_info.orcid else None,
                'affiliation': user_info.affiliation if user_info.affiliation else None
            })
    infos['users'] = user_infos

    locations = logic.locations.get_locations()
    all_relevant_location_ids = set()
    while relevant_location_ids:
        new_relevant_location_ids = set()
        for location_info in locations:
            if location_info.id in relevant_location_ids:
                new_relevant_location_ids.add(location_info.parent_location_id)
        all_relevant_location_ids.update(relevant_location_ids)
        relevant_location_ids = new_relevant_location_ids - all_relevant_location_ids

    location_infos = []
    for location_info in locations:
        if location_info.id in all_relevant_location_ids:
            location_infos.append({
                'id': location_info.id,
                'parent_id': location_info.parent_location_id,
                'name': location_info.name,
                'description': location_info.description
            })
    infos['locations'] = location_infos

    for file_name in relevant_markdown_images:
        content = logic.markdown_images.get_markdown_image(file_name, user_id)
        if content is not None:
            archive_files[f'sampledb_export/markdown_files/{file_name}'] = content

    archive_files['sampledb_export/data.json'] = json.dumps(infos, indent=2)

    readme_text = f"""SampleDB Export
===============

The file data.json contains information on objects from the SampleDB instance at {flask.url_for('frontend.index', _external=True)} in JSON format. Aside from information on the objects themselves, it also contains information on relevant actions, instruments, users and locations.
"""
    if len(archive_files) > 1:
        readme_text += """
The objects directory contains files uploaded for the objects in data.json.
"""
    readme_text += f"""
This archive was created for user #{user_id} at {datetime.datetime.now().isoformat()}.
"""

    archive_files["sampledb_export/README.txt"] = readme_text

    for file_name, file_content in archive_files.items():
        if isinstance(file_content, str):
            archive_files[file_name] = file_content.encode('utf-8')

    return archive_files


def get_zip_archive(user_id: int, object_ids: typing.Optional[typing.List[int]] = None) -> bytes:
    archive_files = get_archive_files(user_id, object_ids=object_ids)
    zip_bytes = io.BytesIO()
    with zipfile.ZipFile(zip_bytes, 'w') as zip_file:
        for file_name, file_content in archive_files.items():
            zip_file.writestr(file_name, file_content)
    return zip_bytes.getvalue()


def get_tar_gz_archive(user_id: int, object_ids: typing.Optional[typing.List[int]] = None) -> bytes:
    archive_files = get_archive_files(user_id, object_ids=object_ids)
    tar_bytes = io.BytesIO()
    with tarfile.open('sampledb_export.tar.gz', 'w:gz', fileobj=tar_bytes) as tar_file:
        for file_name, file_content in archive_files.items():
            tar_info = tarfile.TarInfo(file_name)
            tar_info.size = len(file_content)
            tar_info.mode = 0o444
            tar_info.mtime = time.time()
            tar_file.addfile(tar_info, fileobj=io.BytesIO(file_content))
    return tar_bytes.getvalue()


FILE_FORMATS = {
    '.zip': ('.zip Archive', get_zip_archive, 'application/zip'),
    '.tar.gz': ('.tar.gz Archive', get_tar_gz_archive, 'application/gzip')
}



# coding: utf-8
"""

"""

import os
import flask

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

frontend = flask.Blueprint(
    'frontend',
    __name__,
    template_folder=os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
)


from . import actions
from . import action_types
from . import authentication_forms
from . import export
from . import favicon
from . import index
from . import instruments
from . import locations
from . import markdown_images
from . import objects
from . import objects_forms
from . import projects
from . import status
from . import tags
from . import users
from . import users_forms
from . import errors
from . import utils


__all__ = [
    'actions',
    'action_types',
    'authentication_forms',
    'export',
    'favicon',
    'index',
    'instruments',
    'locations',
    'markdown_images',
    'objects',
    'objects_forms',
    'projects',
    'status',
    'tags',
    'users',
    'users_forms',
    'errors',
    'utils'
]



# coding: utf-8
"""

"""

import datetime

import flask
import flask_login

from .. import frontend
from ...logic.users import create_user, get_user_invitation, set_user_invitation_accepted
from ...logic.authentication import add_email_authentication
from ...logic.utils import send_user_invitation_email
from ...logic.security_tokens import verify_token

from ...models import users
from ...models import Authentication, UserType

from ... import db

from ..users_forms import InvitationForm, RegistrationForm


@frontend.route('/users/invitation', methods=['GET', 'POST'])
def invitation_route():
    if flask_login.current_user.is_authenticated:
        return invitation()
    elif 'token' in flask.request.args:
        return registration()
    else:
        return flask.abort(403)


def invitation():
    invitation_form = InvitationForm()
    if flask.request.method == "GET":
        #  GET (invitation dialog )
        has_success = False
        return flask.render_template('invitation.html', invitation_form=invitation_form, has_success=has_success)
    if flask.request.method == "POST":
        # POST (send invitation)
        has_success = False
        has_error = False
        if invitation_form.validate_on_submit():
            email = invitation_form.email.data
            if '@' not in email:
                has_error = True
            else:
                has_success = True

                user_invitation = users.UserInvitation(
                    inviter_id=flask_login.current_user.id,
                    utc_datetime=datetime.datetime.utcnow()
                )
                db.session.add(user_invitation)
                db.session.commit()

                # send confirm link
                send_user_invitation_email(
                    email=email,
                    invitation_id=user_invitation.id
                )
        else:
            has_error = True
        return flask.render_template('invitation.html', invitation_form=invitation_form, has_success=has_success,
                                     has_error=has_error)


def registration():
    token = flask.request.args.get('token')
    expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
    token_data = verify_token(token, salt='invitation', secret_key=flask.current_app.config['SECRET_KEY'], expiration=expiration_time_limit)
    if token_data is None:
        flask.flash('Invalid invitation token. Please request a new invitation.', 'error')
        return flask.abort(403)
    if isinstance(token_data, str):
        email = token_data
        invitation_id = None
    else:
        email = token_data['email']
        invitation_id = token_data['invitation_id']
        if get_user_invitation(invitation_id).accepted:
            flask.flash('This invitation token has already been used. Please request a new invitation.', 'error')
            return flask.abort(403)
    registration_form = RegistrationForm()
    if registration_form.email.data is None or registration_form.email.data == "":
        registration_form.email.data = email
    has_error = False
    if flask.request.method == "GET":
        # confirmation dialog
        return flask.render_template('registration.html', registration_form=registration_form, has_error=has_error)
    if flask.request.method == "POST":
        # redirect or register user and redirect
        has_error = False
        if registration_form.email.data != email:
            has_error = True
            return flask.render_template('registration.html', registration_form=registration_form, has_error=has_error)
        if registration_form.validate_on_submit():
            name = registration_form.name.data
            password = registration_form.password.data
            # check, if email in authentication-method already exists
            authentication_method = Authentication.query.filter(Authentication.login['login'].astext == email).first()
            # no user with this name and contact email in db => add to db
            if authentication_method is None:
                user = create_user(name=name, email=email, type=UserType.PERSON)
                add_email_authentication(user.id, email, password)
                if invitation_id is not None:
                    set_user_invitation_accepted(invitation_id)
                flask.flash('Your account has been created successfully.', 'success')
                flask_login.login_user(user)
                return flask.redirect(flask.url_for('frontend.index'))
            flask.flash('There already is an account with this email address. Please use this account or contact an administrator.', 'error')
            return flask.redirect(flask.url_for('frontend.sign_in'))
        else:
            return flask.render_template('registration.html', registration_form=registration_form, has_error=has_error)



# coding: utf-8
"""
Script for running the SampleDB server.

Usage: python -m sampledb run [<port>]
"""

import sys
import cherrypy

from .. import create_app


def main(arguments):
    if len(arguments) > 1:
        print(__doc__)
        exit(1)
    if arguments:
        port = arguments[0]
        try:
            port = int(port)
            if port < 1024 or port > 65535:
                raise ValueError()
        except ValueError:
            print("Error: port must be between 1024 and 65535", file=sys.stderr)
            exit(1)
    else:
        port = 8000
    app = create_app()
    cherrypy.tree.graft(app, app.config['SERVER_PATH'])
    cherrypy.config.update({
        'environment': 'production',
        'server.socket_host': '0.0.0.0',
        'server.socket_port': port,
        'server.thread_pool': 4,
        'log.screen': True
    })
    cherrypy.engine.start()
    cherrypy.engine.block()



# coding: utf-8
"""

"""

from . import activity
from . import api_log
from . import authentication
from . import create_other_user
from . import favorites
from . import groups
from . import invitation
from . import list
from . import notifications
from . import preferences
from . import profile


__all__ = [
    'activity',
    'api_log',
    'authentication',
    'create_other_user',
    'favorites',
    'groups',
    'invitation',
    'list',
    'notifications',
    'preferences',
    'profile'
]



# coding: utf-8
"""
This module defines functions which can help filtering objects by using the
sampledb.datetypes types. Each function returns a SQLAlchemy object which can
be used for the filter_func in
VersionedJSONSerializableObjectTables.get_current_objects().

Filters that use equality cannot be expected to be exact due to floating point
precision. These filters use operators that include a range of EPSILON of the
left operand's magnitude in base units.
"""

import operator
import json
import sqlalchemy as db
from . import datatypes

EPSILON = 1e-7

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def float_operator_equals(left, right):
    return db.and_(
        left * (1 - EPSILON) <= right,
        left * (1 + EPSILON) >= right
    )


def float_operator_less_than_equals(left, right):
    return db.and_(
        left * (1 - EPSILON) <= right
    )


def float_operator_greater_than_equals(left, right):
    return db.and_(
        left * (1 + EPSILON) >= right
    )


def quantity_binary_operator(db_obj, other, operator):
    return db.and_(
        db_obj['_type'].astext == 'quantity',
        db_obj['dimensionality'].astext == str(other.dimensionality),
        operator(db_obj['magnitude_in_base_units'].astext.cast(db.Float), other.magnitude_in_base_units)
    )


def quantity_equals(db_obj, other):
    return quantity_binary_operator(db_obj, other, float_operator_equals)


def quantity_less_than(db_obj, other):
    return quantity_binary_operator(db_obj, other, operator.lt)


def quantity_less_than_equals(db_obj, other):
    return quantity_binary_operator(db_obj, other, float_operator_less_than_equals)


def quantity_greater_than(db_obj, other):
    return quantity_binary_operator(db_obj, other, operator.gt)


def quantity_greater_than_equals(db_obj, other):
    return quantity_binary_operator(db_obj, other, float_operator_greater_than_equals)


def quantity_between(db_obj, left, right, including=True):
    if left.dimensionality != right.dimensionality:
        return False
    if including:
        return db.and_(
            db_obj['_type'].astext == 'quantity',
            db_obj['dimensionality'].astext == str(left.dimensionality),
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) * (1 + EPSILON) >= left.magnitude_in_base_units,
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) * (1 - EPSILON) <= right.magnitude_in_base_units
        )
    else:
        return db.and_(
            db_obj['_type'].astext == 'quantity',
            db_obj['dimensionality'].astext == str(left.dimensionality),
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) > left.magnitude_in_base_units,
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) < right.magnitude_in_base_units
        )


def datetime_binary_operator(db_obj, other, operator):
    if isinstance(other, datatypes.DateTime):
        other = other.utc_datetime
    other = other.date()
    return db.and_(
        db_obj['_type'].astext == 'datetime',
        operator(db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD'), other)
    )


def datetime_equals(db_obj, other):
    return datetime_binary_operator(db_obj, other, operator.eq)


def datetime_less_than(db_obj, other):
    return datetime_binary_operator(db_obj, other, operator.lt)


def datetime_less_than_equals(db_obj, other):
    return datetime_binary_operator(db_obj, other, operator.le)


def datetime_greater_than(db_obj, other):
    return datetime_binary_operator(db_obj, other, operator.gt)


def datetime_greater_than_equals(db_obj, other):
    return datetime_binary_operator(db_obj, other, operator.ge)


def datetime_between(db_obj, left, right, including=True):
    if isinstance(left, datatypes.DateTime):
        left = left.utc_datetime
    if isinstance(right, datatypes.DateTime):
        right = right.utc_datetime
    left = left.date()
    right = right.date()
    if including:
        return db.and_(
            db_obj['_type'].astext == 'datetime',
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') >= left,
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') <= right,
        )
    else:
        return db.and_(
            db_obj['_type'].astext == 'datetime',
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') > left,
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') < right,
        )


def boolean_equals(db_obj, value):
    if isinstance(value, datatypes.Boolean):
        value = value.value
    return db.and_(
        db_obj['_type'].astext == 'bool',
        db_obj['value'].astext.cast(db.Boolean) == value
    )


def boolean_true(db_obj):
    return boolean_equals(db_obj, True)


def boolean_false(db_obj):
    return boolean_equals(db_obj, False)


def text_equals(db_obj, text):
    if isinstance(text, datatypes.Text):
        text = text.text
    return db.and_(
        db_obj['_type'].astext == 'text',
        db_obj['text'].astext == text
    )


def text_contains(db_obj, text):
    return db.and_(
        db_obj['_type'].astext == 'text',
        db_obj['text'].astext.like('%' + text + '%')
    )


def sample_equals(db_obj, object_id):
    return db.and_(
        db_obj['_type'].astext == 'sample',
        db_obj['object_id'].astext.cast(db.Integer) == object_id
    )


def tags_contain(db_obj, tag):
    tag = tag.strip().lower()
    return db.and_(
        db_obj['_type'].astext == 'tags',
        db_obj['tags'].contains(json.dumps(tag))
    )



# coding: utf-8
"""
Add data column to files table.
"""

import os
import sqlalchemy

MIGRATION_INDEX = 8
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'files'
    """).fetchall()
    if ('data',) in column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE files
        ADD data JSON
    """)

    files_to_update = []
    for id, object_id, original_file_name in db.session.execute("""SELECT id, object_id, original_file_name FROM files""").fetchall():
        if original_file_name:
            files_to_update.append((id, object_id, original_file_name))
    for id, object_id, original_file_name in files_to_update:
        db.session.execute(sqlalchemy.text("""
        UPDATE files
        SET data=jsonb_set('{"storage": "local"}'::jsonb, '{original_file_name}', to_json(:original_file_name ::text)::jsonb, true), original_file_name=''
        WHERE id=:id AND object_id=:object_id
        """), params={'id': id, 'object_id': object_id, 'original_file_name': original_file_name})

    db.session.execute("""
        ALTER TABLE files
        DROP COLUMN original_file_name
    """)
    return True



# coding: utf-8
"""
Script for listing all instruments in SampleDB.

Usage: python -m sampledb list_instruments
"""

from .. import create_app
from ..logic.instruments import get_instruments


def main(arguments):
    if len(arguments) != 0:
        print(__doc__)
        exit(1)
    app = create_app()
    with app.app_context():
        instruments = get_instruments()
        for instrument in instruments:
            print(" - #{0.id}: {0.name}".format(instrument))



# coding: utf-8
"""
RESTful API for SampleDB
"""

import flask

from flask_restful import Resource

from .authentication import object_permissions_required, Permissions
from ...logic.comments import Comment, create_comment, get_comments_for_object, get_comment_for_object
from ...logic import errors

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def comment_to_json(comment: Comment):
    comment_json = {
        'object_id': comment.object_id,
        'comment_id': comment.id,
        'user_id': comment.user_id,
        'utc_datetime': comment.utc_datetime.isoformat(),
        'content': comment.content
    }
    return comment_json


class ObjectComment(Resource):
    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int, comment_id: int):
        try:
            comment = get_comment_for_object(object_id=object_id, comment_id=comment_id)
        except errors.CommentDoesNotExistError:
            return {
                "message": "comment {} of object {} does not exist".format(comment_id, object_id)
            }, 404
        return comment_to_json(comment)


class ObjectComments(Resource):
    @object_permissions_required(Permissions.WRITE)
    def post(self, object_id: int):
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        if 'object_id' in request_json:
            if request_json['object_id'] != object_id:
                return {
                    "message": "object_id must be {}".format(object_id)
                }, 400
        if 'content' not in request_json:
            return {
                "message": "content must be set"
            }, 400
        content = request_json['content']
        if not isinstance(content, str):
            return {
                "message": "content must be a string"
            }, 400
        if not content:
            return {
                "message": "content must not be empty"
            }, 400
        comment_id = create_comment(object_id, flask.g.user.id, content)

        comment_url = flask.url_for(
            'api.object_comment',
            object_id=object_id,
            comment_id=comment_id
        )
        return flask.redirect(comment_url, code=201)

    @object_permissions_required(Permissions.READ)
    def get(self, object_id: int):
        return [
            comment_to_json(comment)
            for comment in get_comments_for_object(object_id)
        ]



# coding: utf-8
"""
Add user_id column to actions table.
"""

import os

MIGRATION_INDEX = 1
MIGRATION_NAME, _ = os.path.splitext(os.path.basename(__file__))


def run(db):
    # Skip migration by condition
    client_column_names = db.session.execute("""
        SELECT column_name
        FROM information_schema.columns
        WHERE table_name = 'actions'
    """).fetchall()
    if ('user_id',) in client_column_names:
        return False

    # Perform migration
    db.session.execute("""
        ALTER TABLE actions
        ADD user_id INTEGER
    """)
    db.session.execute("""
        ALTER TABLE actions
        ADD FOREIGN KEY(user_id) REFERENCES users(id)
    """)
    return True



# coding: utf-8
import functools
import json
import typing
from sqlalchemy import String, and_, or_
from sqlalchemy.sql.expression import select, true, false, not_
from . import where_filters
from . import datatypes
from . import object_search_parser
from .. import db
import sqlalchemy.dialects.postgresql as postgresql


class Attribute:
    def __init__(self, input_text: str, start_position: int, value):
        self.value = value
        self.input_text = input_text
        self.start_position = start_position
        self.end_position = self.start_position + len(self.input_text)


class Expression:
    def __init__(self, input_text: str, start_position: int, value):
        self.value = value
        self.input_text = input_text
        self.start_position = start_position
        self.end_position = self.start_position + len(self.input_text)


unary_operator_handlers = {}
binary_operator_handlers = {}


def unary_operator_handler(operand_type, operator):
    def unary_operator_handler_decorator(func, operand_type=operand_type, operator=operator):
        @functools.wraps(func)
        def unary_operator_handler_wrapper(operator, operand, outer_filter, search_notes):
            input_text = operator.input_text + operand.input_text
            start_position = operator.start_position
            end_position = operand.end_position
            filter_func, outer_filter = func(operand.value, outer_filter, search_notes, input_text, start_position, end_position)
            filter_func = Expression(input_text, start_position, filter_func)
            return filter_func, outer_filter
        assert (operand_type, operator) not in unary_operator_handlers
        unary_operator_handlers[(operand_type, operator)] = unary_operator_handler_wrapper
        return unary_operator_handler_wrapper

    return unary_operator_handler_decorator


def binary_operator_handler(left_operand_type, right_operand_type, operator):
    def binary_operator_handler_decorator(func, left_operand_type=left_operand_type, right_operand_type=right_operand_type, operator=operator):
        @functools.wraps(func)
        def binary_operator_handler_wrapper(left_operand, operator, right_operand, outer_filter, search_notes):
            input_text = left_operand.input_text + operator.input_text + right_operand.input_text
            start_position = left_operand.start_position
            end_position = right_operand.end_position
            filter_func, outer_filter = func(left_operand.value, right_operand.value, outer_filter, search_notes, input_text, start_position, end_position)
            filter_func = Expression(input_text, start_position, filter_func)
            return filter_func, outer_filter
        assert (left_operand_type, right_operand_type, operator) not in binary_operator_handlers
        binary_operator_handlers[(left_operand_type, right_operand_type, operator)] = binary_operator_handler_wrapper
        return binary_operator_handler_wrapper

    return binary_operator_handler_decorator


@unary_operator_handler(datatypes.Boolean, 'not')
def _(operand, outer_filter, search_notes, input_text, start_position, end_position):
    if not operand.value:
        search_notes.append(('warning', 'This expression will always be true', start_position, end_position))
        return outer_filter(true()), None
    else:
        search_notes.append(('warning', 'This expression will always be false', start_position, end_position))
        return outer_filter(false()), None


@unary_operator_handler(Attribute, 'not')
def _(operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.boolean_false(operand)), None


@unary_operator_handler(None, 'not')
def _(operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(not_(operand)), None


@binary_operator_handler(datatypes.Boolean, datatypes.Boolean, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value and right_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(None, None, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(and_(left_operand, right_operand)), None


@binary_operator_handler(None, datatypes.Boolean, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if right_operand.value:
        return outer_filter(left_operand), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Boolean, None, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value:
        return outer_filter(right_operand), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(None, Attribute, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(and_(left_operand, where_filters.boolean_true(right_operand))), None


@binary_operator_handler(Attribute, None, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(and_(where_filters.boolean_true(left_operand), right_operand)), None


@binary_operator_handler(Attribute, datatypes.Boolean, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if right_operand.value:
        return outer_filter(where_filters.boolean_true(left_operand)), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Boolean, Attribute, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value:
        return outer_filter(where_filters.boolean_true(right_operand)), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(Attribute, Attribute, 'and')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(and_(where_filters.boolean_true(left_operand), where_filters.boolean_true(right_operand))), None


@binary_operator_handler(datatypes.Boolean, datatypes.Boolean, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value or right_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(None, None, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(or_(left_operand, right_operand)), None


@binary_operator_handler(None, datatypes.Boolean, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if right_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(left_operand), None


@binary_operator_handler(datatypes.Boolean, None, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(right_operand), None


@binary_operator_handler(None, Attribute, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(or_(left_operand, where_filters.boolean_true(right_operand))), None


@binary_operator_handler(Attribute, None, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(or_(where_filters.boolean_true(left_operand), right_operand)), None


@binary_operator_handler(Attribute, datatypes.Boolean, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if right_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(where_filters.boolean_true(left_operand)), None


@binary_operator_handler(datatypes.Boolean, Attribute, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.value:
        return outer_filter(true()), None
    else:
        return outer_filter(where_filters.boolean_true(right_operand)), None


@binary_operator_handler(Attribute, Attribute, 'or')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(or_(where_filters.boolean_true(left_operand), where_filters.boolean_true(right_operand))), None


@binary_operator_handler(datatypes.Boolean, datatypes.Boolean, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand == right_operand:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand == right_operand:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Text, datatypes.Text, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand == right_operand:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(false()), None
    if left_operand == right_operand:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Boolean, datatypes.Boolean, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if not (left_operand == right_operand):
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if not (left_operand == right_operand):
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Text, datatypes.Text, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if not (left_operand == right_operand):
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(true()), None
    if not (left_operand == right_operand):
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Boolean, Attribute, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.boolean_equals(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.Boolean, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.boolean_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Boolean, Attribute, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.boolean_equals(right_operand, not left_operand.value)), None


@binary_operator_handler(Attribute, datatypes.Boolean, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.boolean_equals(left_operand, not right_operand.value)), None


@binary_operator_handler(Attribute, datatypes.DateTime, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.DateTime, Attribute, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_equals(right_operand, left_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_equals(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.Quantity, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(not_(where_filters.quantity_equals(right_operand, left_operand))), None


@binary_operator_handler(Attribute, datatypes.Quantity, '!=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(not_(where_filters.quantity_equals(left_operand, right_operand))), None


@binary_operator_handler(datatypes.Text, Attribute, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.text_equals(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.Text, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.text_equals(left_operand, right_operand)), None


@binary_operator_handler(Attribute, Attribute, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(left_operand == right_operand), None


@binary_operator_handler(None, None, '==')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(left_operand == right_operand), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '<')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.utc_datetime < right_operand.utc_datetime:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '<')
def _(left_operand: datatypes.Quantity, right_operand: datatypes.Quantity, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(false()), None
    if left_operand.magnitude_in_base_units < right_operand.magnitude_in_base_units:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(Attribute, datatypes.Quantity, '<')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_less_than(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '<')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_greater_than(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.DateTime, '<')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_less_than(left_operand, right_operand)), None


@binary_operator_handler(datatypes.DateTime, Attribute, '<')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_greater_than(right_operand, left_operand)), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '>')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.utc_datetime > right_operand.utc_datetime:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '>')
def _(left_operand: datatypes.Quantity, right_operand: datatypes.Quantity, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(false()), None
    if left_operand.magnitude_in_base_units > right_operand.magnitude_in_base_units:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(Attribute, datatypes.Quantity, '>')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_greater_than(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '>')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_less_than(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.DateTime, '>')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_greater_than(left_operand, right_operand)), None


@binary_operator_handler(datatypes.DateTime, Attribute, '>')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_less_than_equals(right_operand, left_operand)), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '<=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.utc_datetime <= right_operand.utc_datetime:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '<=')
def _(left_operand: datatypes.Quantity, right_operand: datatypes.Quantity, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(false()), None
    if left_operand.magnitude_in_base_units <= right_operand.magnitude_in_base_units:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(Attribute, datatypes.Quantity, '<=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_less_than_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '<=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_greater_than_equals(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.DateTime, '<=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_less_than_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.DateTime, Attribute, '<=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_greater_than_equals(right_operand, left_operand)), None


@binary_operator_handler(datatypes.DateTime, datatypes.DateTime, '>=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.utc_datetime >= right_operand.utc_datetime:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(datatypes.DateTime, Attribute, '>=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_less_than_equals(right_operand, left_operand)), None


@binary_operator_handler(datatypes.Quantity, datatypes.Quantity, '>=')
def _(left_operand: datatypes.Quantity, right_operand: datatypes.Quantity, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.dimensionality != right_operand.dimensionality:
        search_notes.append(('warning', 'Invalid comparison between quantities of different dimensionalities', 0, None))
        return outer_filter(false()), None
    if left_operand.magnitude_in_base_units >= right_operand.magnitude_in_base_units:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


@binary_operator_handler(Attribute, datatypes.Quantity, '>=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_greater_than_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Quantity, Attribute, '>=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.quantity_less_than_equals(right_operand, left_operand)), None


@binary_operator_handler(Attribute, datatypes.DateTime, '>=')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.datetime_greater_than_equals(left_operand, right_operand)), None


@binary_operator_handler(datatypes.Text, Attribute, 'in')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    return outer_filter(where_filters.text_contains(right_operand, left_operand.text)), None


@binary_operator_handler(datatypes.Text, datatypes.Text, 'in')
def _(left_operand, right_operand, outer_filter, search_notes, input_text, start_position, end_position):
    if left_operand.text in right_operand.text:
        return outer_filter(true()), None
    else:
        return outer_filter(false()), None


def transform_literal_to_query(data, literal: object_search_parser.Literal, search_notes: typing.List[typing.Tuple[str, str, int, typing.Optional[int]]]) -> typing.Tuple[typing.Any, typing.Optional[typing.Callable]]:
    if isinstance(literal, object_search_parser.Tag):
        return Expression(literal.input_text, literal.start_position, where_filters.tags_contain(data[('tags',)], literal.value)), None

    if isinstance(literal, object_search_parser.Attribute):
        attributes = literal.value
        # covert any numeric arguments to integers (array indices)
        for i, attribute in enumerate(attributes):
            try:
                attributes[i] = int(attribute)
            except ValueError:
                pass
        if '?' in attributes:
            array_placeholder_index = attributes.index('?')
            # no danger of SQL injection as attributes may only consist of
            # characters and underscores at this point
            jsonb_selector = ''
            for i, attribute in enumerate(attributes[:array_placeholder_index]):
                if isinstance(attribute, int):
                    attribute = str(attribute)
                else:
                    attribute = "'" + attribute + "'"
                if i > 0:
                    jsonb_selector += " -> "
                jsonb_selector += attribute
            array_items = select(columns=[db.text('value FROM jsonb_array_elements_text(data -> {})'.format(jsonb_selector))])
            db_obj = db.literal_column('value').cast(postgresql.JSONB)
            for attribute in attributes[array_placeholder_index + 1:]:
                db_obj = db_obj[attribute]
            return Attribute(literal.input_text, literal.start_position, db_obj), lambda filter: db.Query(db.literal(True)).select_entity_from(array_items).filter(filter).exists()
        return Attribute(literal.input_text, literal.start_position, data[attributes]), None

    if isinstance(literal, object_search_parser.Boolean):
        return literal, None

    if isinstance(literal, object_search_parser.Date):
        return literal, None

    if isinstance(literal, object_search_parser.Quantity):
        return literal, None

    if isinstance(literal, object_search_parser.Text):
        return literal, None

    search_notes.append(('error', "Invalid search query", 0, None))
    return false(), None


def transform_unary_operation_to_query(data, operator, operand, search_notes: typing.List[typing.Tuple[str, str, int, typing.Optional[int]]]) -> typing.Tuple[Expression, typing.Optional[typing.Callable]]:
    start_token = operator
    start = start_token.start_position
    end_token = operand
    while isinstance(end_token, list):
        end_token = end_token[0]
    end = end_token.start_position + len(end_token.input_text)
    operand, outer_filter = transform_tree_to_query(data, operand, search_notes)
    if not outer_filter:
        def outer_filter(filter):
            return filter

    str_operator = operator.operator
    operator_aliases = {
        '!': 'not'
    }
    if str_operator in operator_aliases:
        str_operator = operator_aliases[str_operator]

    if isinstance(operand, object_search_parser.Boolean):
        operand_type = datatypes.Boolean
    elif isinstance(operand, object_search_parser.Date):
        operand_type = datatypes.DateTime
    elif isinstance(operand, object_search_parser.Quantity):
        operand_type = datatypes.Quantity
    elif isinstance(operand, object_search_parser.Text):
        operand_type = datatypes.Text
    elif isinstance(operand, Attribute):
        operand_type = Attribute
    else:
        operand_type = None

    if (operand_type, str_operator) in unary_operator_handlers:
        return unary_operator_handlers[(operand_type, str_operator)](operator, operand, outer_filter, search_notes)

    search_notes.append(('error', "Unknown unary operation", start, end))
    return Expression(operator.input_text + operand.input_text, operator.start_position, false()), None


def transform_binary_operation_to_query(data, left_operand, operator, right_operand, search_notes: typing.List[typing.Tuple[str, str, int, typing.Optional[int]]]) -> typing.Tuple[Expression, typing.Optional[typing.Callable]]:
    start_token = left_operand
    while isinstance(start_token, list):
        start_token = start_token[0]
    start = start_token.start_position
    end_token = right_operand
    while isinstance(end_token, list):
        end_token = end_token[-1]
    end = end_token.start_position + len(end_token.input_text)

    left_operand, left_outer_filter = transform_tree_to_query(data, left_operand, search_notes)
    right_operand, right_outer_filter = transform_tree_to_query(data, right_operand, search_notes)

    if left_outer_filter and right_outer_filter:
        search_notes.append(('error', "Multiple array placeholders", start, end))
        return Expression(left_operand.input_text + operator.input_text + right_operand.input_text, left_operand.start_position, false()), None
    if left_outer_filter:
        outer_filter = left_outer_filter
    elif right_outer_filter:
        outer_filter = right_outer_filter
    else:
        def outer_filter(filter):
            return filter

    str_operator = operator.operator
    operator_aliases = {
        '||': 'or',
        '&&': 'and',
        '=': '=='
    }
    if str_operator in operator_aliases:
        str_operator = operator_aliases[str_operator]

    if isinstance(left_operand, object_search_parser.Boolean):
        left_operand_type = datatypes.Boolean
    elif isinstance(left_operand, object_search_parser.Date):
        left_operand_type = datatypes.DateTime
    elif isinstance(left_operand, object_search_parser.Quantity):
        left_operand_type = datatypes.Quantity
    elif isinstance(left_operand, object_search_parser.Text):
        left_operand_type = datatypes.Text
    elif isinstance(left_operand, Attribute):
        left_operand_type = Attribute
    else:
        left_operand_type = None

    if isinstance(right_operand, object_search_parser.Boolean):
        right_operand_type = datatypes.Boolean
    elif isinstance(right_operand, object_search_parser.Date):
        right_operand_type = datatypes.DateTime
    elif isinstance(right_operand, object_search_parser.Quantity):
        right_operand_type = datatypes.Quantity
    elif isinstance(right_operand, object_search_parser.Text):
        right_operand_type = datatypes.Text
    elif isinstance(right_operand, Attribute):
        right_operand_type = Attribute
    else:
        right_operand_type = None

    if datatypes.DateTime in (left_operand_type, right_operand_type):
        if str_operator.strip() == 'after':
            str_operator = '>'
        elif str_operator.strip() == 'before':
            str_operator = '<'
        elif str_operator.strip() == 'on':
            str_operator = '=='

    if (left_operand_type, right_operand_type, str_operator) in binary_operator_handlers:
        return binary_operator_handlers[(left_operand_type, right_operand_type, str_operator)](left_operand, operator, right_operand, outer_filter, search_notes)

    search_notes.append(('error', "Unknown binary operation", start, end))
    return Expression(left_operand.input_text + operator.input_text + right_operand.input_text, left_operand.start_position, false()), None


def transform_tree_to_query(data, tree: typing.Union[object_search_parser.Literal, list], search_notes: typing.List[typing.Tuple[str, str, int, typing.Optional[int]]]) -> typing.Tuple[Expression, typing.Optional[typing.Callable]]:
    if isinstance(tree, object_search_parser.Literal):
        return transform_literal_to_query(data, tree, search_notes)
    if not isinstance(tree, list):
        search_notes.append(('error', "Invalid search query", 0, None))
        return false(), None
    if len(tree) == 1:
        value, = tree
        return transform_tree_to_query(data, value, search_notes)
    if len(tree) == 2:
        operator, operand = tree
        if isinstance(operator, object_search_parser.Operator):
            return transform_unary_operation_to_query(data, operator, operand, search_notes)
        search_notes.append(('error', "Invalid search query (missing operator)", 0, None))
        return false(), None
    if len(tree) == 3:
        left_operand, operator, right_operand = tree
        if isinstance(operator, object_search_parser.Operator):
            return transform_binary_operation_to_query(data, left_operand, operator, right_operand, search_notes)
        search_notes.append(('error', "Invalid search query (missing operator)", 0, None))
        return false(), None


def should_use_advanced_search(query_string: str) -> typing.Tuple[bool, str]:
    """
    Detect whether the advanced search should be used automatically.

    The user can force the use of the advanced search, but for specific search
    query strings, the advanced search will be more appropriate, e.g. when
    the user tries to use comparisons or search for tags.

    To prevent an automatic advanced search, users can quote their query
    strings. This way they will use the simple, text-based search.

    :param query_string: the original query string
    :return: whether to use the advanced search and a modified query string
    """
    if query_string[0] == query_string[-1] == '"':
        # Remove quotes around the query string
        return False, query_string[1:-1]

    for operator in ('=', '<', '>', '#', '&', '|'):
        if operator in query_string:
            return True, query_string

    return False, query_string


def generate_filter_func(query_string: str, use_advanced_search: bool) -> typing.Tuple[typing.Callable, typing.Any, bool]:
    """
    Generates a filter function for use with SQLAlchemy and the JSONB data
    attribute in the object tables.

    The generated filter functions can be used for objects.get_objects()

    :param query_string: the query string
    :param use_advanced_search: whether to use simple text search (False) or advanced search (True)
    :return: filter func, search tree and whether the advanced search was used
    """
    tree = None
    query_string = query_string.strip()
    if query_string:
        if not use_advanced_search:
            use_advanced_search, query_string = should_use_advanced_search(query_string)
        if use_advanced_search:
            # Advanced search using parser and where_filters
            try:
                tree = object_search_parser.parse_query_string(query_string)
            except object_search_parser.ParseError as e:
                def filter_func(data, search_notes, e=e):
                    """ Return no objects and set search_notes"""
                    search_notes.append(('error', e.message, e.start, e.end))
                    return False
                return filter_func, None, use_advanced_search
            except Exception:
                def filter_func(data, search_notes, start=0, end=len(query_string)):
                    """ Return no objects and set search_notes"""
                    search_notes.append(('error', "Failed to parse query string", start, end))
                    return False
                return filter_func, None, use_advanced_search
            if isinstance(tree, list) and not tree:
                def filter_func(data, search_notes, start=0, end=len(query_string)):
                    """ Return no objects and set search_notes"""
                    search_notes.append(('error', 'Empty search', start, end))
                    return False
                return filter_func, None, use_advanced_search
            if isinstance(tree, object_search_parser.Literal):
                if isinstance(tree, object_search_parser.Boolean):
                    if tree.value.value:
                        def filter_func(data, search_notes, start=0, end=len(query_string)):
                            """ Return all objects and set search_notes"""
                            search_notes.append(('warning', 'This search will always return all objects', start, end))
                            return True
                        return filter_func, tree, use_advanced_search
                    else:
                        def filter_func(data, search_notes, start=0, end=len(query_string)):
                            """ Return no objects and set search_notes"""
                            search_notes.append(('warning', 'This search will never return any objects', start, end))
                            return False
                        return filter_func, tree, use_advanced_search
                elif isinstance(tree, object_search_parser.Attribute):
                    pass
                elif isinstance(tree, object_search_parser.Tag):
                    pass
                else:
                    def filter_func(data, search_notes, start=0, end=len(query_string)):
                        """ Return no objects and set search_notes"""
                        search_notes.append(('error', 'Unable to use literal as search query', start, end))
                        return False
                    return filter_func, None, use_advanced_search

            def filter_func(data, search_notes, tree=tree):
                """ Filter objects based on search query string """
                filter_func, outer_filter = transform_tree_to_query(data, tree, search_notes)
                # check bool if filter_func is only an attribute
                if isinstance(filter_func, Expression):
                    filter_func = filter_func.value
                if isinstance(filter_func, Attribute):
                    filter_func = where_filters.boolean_true(filter_func.value)
                if outer_filter:
                    filter_func = outer_filter(filter_func)
                return filter_func
        else:
            # Simple search in values
            def filter_func(data, search_notes, query_string=query_string):
                """ Filter objects based on search query string """
                # The query string is converted to json to escape quotes, backslashes, etc
                query_string = json.dumps(query_string)[1:-1]
                return data.cast(String).ilike('%: "%' + query_string + '%"%')
    else:
        def filter_func(data, search_notes):
            """ Return all objects"""
            return True
    return filter_func, tree, use_advanced_search


def wrap_filter_func(filter_func):
    """
    Wrap a filter function so that a new list will be filled with the search notes.

    :param filter_func: the filter function to wrap
    :return: the wrapped filter function and the search notes list
    """
    search_notes = []

    def wrapped_filter_func(*args, search_notes=search_notes, filter_func_impl=filter_func, **kwargs):
        return filter_func_impl(*args, search_notes=search_notes, **kwargs)

    return wrapped_filter_func, search_notes
